(define-package "magit" "2.90.1"
  "A Git porcelain inside Emacs."
  '((emacs "25.1")
    (async "20180527")
    (dash "20180910")
    (git-commit "20181104")
    (transient "20190812")
    (with-editor "20181103"))
  :keywords
  '("git" "tools" "vc"))
